/*
 * SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#pragma once

#include "gpu/gpu.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Minimal NOCAT eventbuffer producer interface.
 * Mirrors FECS/Video patterns with a reduced binding API suitable for
 * Linux-only notification via EventBuffer.
 */

struct RsClient;
struct RsResourceRef;
typedef struct EventBuffer EventBuffer;

/* Create a bindpoint between an EventBuffer and NOCAT for this GPU */
NV_STATUS nocatAddBindpoint(
    OBJGPU *pGpu,
    struct RsClient *pClient,
    struct RsResourceRef *pEventBufferRef,
    NvHandle hNotifier);

/* Remove all bindpoints associated with a given EventBuffer (called at destruction). */
void nocatRemoveAllBindpoints(EventBuffer *pEventBuffer);

/* Remove all bindpoints for a GPU (optional cleanup helper). */
void nocatRemoveAllBindpointsForGpu(OBJGPU *pGpu);

/* Produce a minimal NOCAT notification record to all bound buffers on this GPU. */
void nocatEventBufferAddNotify(
    OBJGPU *pGpu,
    NvU32 posted,
    NvU32 recType,
    NvU64 timeStamp);

#ifdef __cplusplus
}
#endif


